# author: abaj
# date: 31.08.2020
# (c) AMS AG
# XML generator for digilevels file in Digital Pattern Editor

from xml.dom import minidom
import os
#import lxml.etree as etree
import xml.dom.minidom
import xml.etree.ElementTree as etr
from xml.dom.minidom import parse
import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)

#PinLevelSheet class is used as parrent class to DigitalPinLevelSets and DcPowerLevelSets
#It has not really class-attributes but only those needed to make an XML doc
#Some of those are :
#root specifies the root of the XML document
#xml specifies the subroot of the XML document and we connect each child/node to him
class PinLevelSheet:
    # XMLHeader stuff, always like this, nothing special

    def __init__(self):
        logger = logging.getLogger("digilevelsSource")
        logger.debug("Building up an XML framework header")
        self.root = minidom.Document()
        self.xml = self.root.createElement('PinLevelsFile')
        self.xml.setAttribute('xmlns:xsd', 'http://www.w3.org/2001/XMLSchema')
        self.xml.setAttribute('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema')
        self.xml.setAttribute('schemaVersion', '1.0')
        self.xml.setAttribute('xmlns', 'http://www.ni.com/Semiconductor/PinLevels')
        self.root.appendChild(self.xml)
        self.pinlevelsheetChild = self.root.createElement('PinLevelsSheet')
        self.xml.appendChild(self.pinlevelsheetChild)
        self.pinType = self.root.createElement('DigitalPinLevelSets')
        self.pinlevelsheetChild.appendChild(self.pinType)
        logger.debug("Building up an XML framework header finished")

        self.dpLevelSets = []
        self.dcLevelSets = []
    
     
    #def createConsoleHandler(self):
    #    self.consoleHandler = logging.StreamHandler()

    #Wrapper function to call the function in the DigitalPinLevelSet abstract class
    def addDigitalPinLevelSet(self, name, Vil, Vih, Vol, Voh, Iol, Ioh, Vcom, Vterm, TerminationMode, Comment):
        """Add a Digital Pin Level Set to the sheet

        Args:
            name (str): Name of the Pin Level Set
            Vil (float):  Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic low (0). 
            Vih (float): Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic high (1). 
            Vol (float): Specifies the DUT output voltage below which the comparator on the instrument pin interprets a logic low (L). 
            Voh (float): Specifies the DUT output voltage above which the comparator on the instrument pin interprets a logic high (H).
            Iol (float): Specifies the current that the DUT sinks from the active load while outputting a voltage below Vcom
            Ioh (float): Specifies the current that the DUT sources to the active load while outputting a voltage above Vcom
            Vcom (float): Specifies the commutating voltage level at which the active load circuit switches between sourcing current and sinking current.
            Vterm (float): Specifies the termination voltage the instrument applies during non-drive cycles when the termination mode is set to VTERM. The instrument applies the termination voltage through a 50 Ω parallel termination resistance.
            TerminationMode (str): Specifies the behavior of the pin when the pin function is set to digital and the pin state for the current cycle is not a drive state. Possible Values: 'ActiveLoad', 'Vterm', 'HighZ'
            Comment (str): comment for the file (optional)
        """
        dplevelset = DigitalPinLevelSet(name, Vil, Vih, Vol, Voh, Iol, Ioh, Vcom, Vterm, TerminationMode, Comment)
        self.dpLevelSets.append(dplevelset)

    #Wrapper function to call the function in the DCPowerLevelSet abstract class
    def addDCPowerLevelsSet(self, name, Vf, If, Comment):
        """Add a DC Power Levels set to the sheet

        Args:
            name (string): Level set Name
            Vf (float): Forward Voltage
            If (float): Forward Current
            Comment (string): Comment
        """
        dcpowerlevelset = DCPowerLevelsSet(name, Vf, If, Comment)
        self.dcLevelSets.append(dcpowerlevelset)

    #Function used to generate XML digilevel file 
    #Parameters:
    #@self specifies the instance of the class which actually generates its XML file
    #@name_of_file specifies the name of the file that we want to generate
    def generateXML(self, name_of_file = ""):
        """generates and writes the xml file for the created PinLevelSheet Object

        Args:
            name_of_file (str): specify the name/path of the digilevels file. Skipping if no name is given 
        """
        if(name_of_file != "" and name_of_file[-11:] == ".digilevels"):
            logger.debug("building xml")

            for dpLevelSet in self.dpLevelSets:
                dpLevelSet.createDigitalPinLevelSets(self.root, self.pinType)

            for dcLevelSet in self.dcLevelSets:
                dcLevelSet.createDCPowerLevelsSets(self.root)

            logger.debug("Writing the digilevels file file")
            save_file = "%s"%(name_of_file)
            with open(save_file, "w") as f:
                self.root.writexml(f, "\n", "  ", encoding="utf-8")

# Class needed to make the Digital Pin Level Sets
# Attributes:
# name - representing the name of the channel
# Vil - Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic low (0).
# Vih - Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic high (1).
# Vol - Specifies the DUT output voltage below which the comparator on the instrument pin interprets a logic low (L).
# Voh - Specifies the DUT output voltage above which the comparator on the instrument pin interprets a logic high (H).
# Vcom - Specifies the commutating voltage level at which the active load circuit switches between sourcing current and sinking current.
# Vterm - Specifies the termination voltage the instrument applies during non-drive cycles when the termination mode is set to VTERM.
#      - The instrument applies the termination voltage through a 50 Ω parallel termination resistance.
# Iol - Specifies the current that the DUT sinks from the active load while outputting a voltage below VCOM.
# Ioh - Specifies the current that the DUT sources to the active load while outputting a voltage above VCOM.
class DigitalPinLevelSet():
    name = " "
    Vil = " "
    Vih = " "
    Vol = " "
    Voh = " "
    Vcom = " "
    Vterm = " "
    Iol = " "
    Ioh = " "
    TerminationMode = " "
    Comment = " "

    #Constructor of this class
    #It sets the value of each attribute, when the instance of the class is made/created
    def __init__(self, name, Vil, Vih, Vol, Voh, Iol, Ioh, Vcom, Vterm, TerminationMode, Comment):
        """Create and initialize Values of a Digital Pin Level Set

        Args:
            name (str): Name of the Pin Level Set
            Vil (float):  Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic low (0). 
            Vih (float): Specifies the voltage that the instrument applies to the input of the DUT when the instrument drives a logic high (1). 
            Vol (float): Specifies the DUT output voltage below which the comparator on the instrument pin interprets a logic low (L). 
            Voh (float): Specifies the DUT output voltage above which the comparator on the instrument pin interprets a logic high (H).
            Iol (float): Specifies the current that the DUT sinks from the active load while outputting a voltage below Vcom
            Ioh (float): Specifies the current that the DUT sources to the active load while outputting a voltage above Vcom
            Vcom (float): Specifies the commutating voltage level at which the active load circuit switches between sourcing current and sinking current.
            Vterm (float): Specifies the termination voltage the instrument applies during non-drive cycles when the termination mode is set to VTERM. The instrument applies the termination voltage through a 50 Ω parallel termination resistance.
            TerminationMode (str): Specifies the behavior of the pin when the pin function is set to digital and the pin state for the current cycle is not a drive state. Possible Values: 'ActiveLoad', 'Vterm', 'HighZ'
            Comment (str): comment for the file (optional)
        """
        self.name = str(name)
        self.Vil = str(Vil)
        self.Vih = str(Vih)
        self.Vol = str(Vol)
        self.Voh = str(Voh)
        self.Iol = str(Iol)
        self.Ioh = str(Ioh)
        self.Vcom = str(Vcom)
        self.Vterm = str(Vterm)
        self.TerminationMode = str(TerminationMode)
        self.Comment = str(Comment)
        #self.createDigitalPinLevelSets()

    #Function to get the info about components of the pin
    def printPin(self):
        """
        prints the infos to the Digital Pin Level Set
        """
        print("PinName  Vil  Vih  Vol   Voh   Iol   Ioh   Vcom  Vterm TerminationMode  Comment""\n", self.name, "   |",
              self.Vil, "|",
              self.Vih, "|",
              self.Vol, "|",
              self.Voh, "|",
              self.Iol, "|",
              self.Ioh, "|",
              self.Vcom, "  |",
              self.Vterm, "  |",
              self.TerminationMode, "|",
              self.Comment, " |")
    
    #Function used to create the Digital Pin Level Sets with the Components from its class.
    #Parameters:
    #@self specifies the instance of the class at which we are currently calling the function
    def createDigitalPinLevelSets(self, root, pinType):
        """appends the Digital Pin level to the pin level sheet

        Args:
            root (PinLevelSheet): the pin level sheet to append to
            pinType ([type]): [description]
        """
        logger.debug("Adding the DigitalPinLevelSet component to the root of the XML doc.")
        digitalPinLevelSet = root.createElement('DigitalPinLevelSet')
        logger.debug("Adding the DigitalPinLevelSet component to the root of the XML doc DONE!")
        digitalPinLevelSet.setAttribute('pin', '%s' % (self.name))
        pinType.appendChild(digitalPinLevelSet)
        Vil = root.createElement('Vil')
        Vil_atr = root.createTextNode("%s" % (self.Vil))
        Vil.appendChild(Vil_atr)
        Vih = root.createElement('Vih')
        Vih_atr = root.createTextNode("%s" % (self.Vih))
        Vih.appendChild(Vih_atr)
        Vol = root.createElement('Vol')
        Vol_atr = root.createTextNode("%s" % (self.Vol))
        Vol.appendChild(Vol_atr)
        Voh = root.createElement('Voh')
        Voh_atr = root.createTextNode("%s" % (self.Voh))
        Voh.appendChild(Voh_atr)
        Iol = root.createElement('Iol')
        Iol_atr = root.createTextNode("%s m" % (self.Iol))
        Iol.appendChild(Iol_atr)
        Ioh = root.createElement('Ioh')
        Ioh_atr = root.createTextNode("%s m" % (self.Ioh))
        Ioh.appendChild(Ioh_atr)
        Vcom = root.createElement('Vcom')
        Vcom_atr = root.createTextNode("%s" % (self.Vcom))
        Vcom.appendChild(Vcom_atr)
        Vterm = root.createElement('Vterm')
        Vterm_atr = root.createTextNode("%s" % (self.Vterm))
        Vterm.appendChild(Vterm_atr)
        TerminationMode = root.createElement('TerminationMode')
        TerminationMode_atr = root.createTextNode("%s" % (self.TerminationMode))
        TerminationMode.appendChild(TerminationMode_atr)
        Comment = root.createElement('Comment')
        Comment_atr = root.createTextNode("%s" % (self.Comment))
        Comment.appendChild(Comment_atr)
        if(convNumbIsValid(exNumFromString(self.Vil))):
            if(len(self.Vil) > 0 and not self.Vil.isspace()):
                digitalPinLevelSet.appendChild(Vil)
            else:
                logger.debug("Invalid move, Vil cannot be made")
        if(convNumbIsValid(exNumFromString(self.Vih))):        
            if(len(self.Vih) > 0 and not self.Vih.isspace()):
                digitalPinLevelSet.appendChild(Vih)
            else:
                logger.debug("Invalid move, Vih cannot be made")
        if(convNumbIsValid(exNumFromString(self.Vol))):
            if(len(self.Vol) > 0 and not self.Vol.isspace()):
                digitalPinLevelSet.appendChild(Vol)
            else:
                logger.debug("Invalid move, Vol cannot be made")
        if(convNumbIsValid(exNumFromString(self.Voh))):
            if(len(self.Voh) > 0 and not self.Voh.isspace()):
                digitalPinLevelSet.appendChild(Voh)
            else:
                logger.debug("Invalid move, Voh cannot be made")
        if(convNumbIsValid(exNumFromString(self.Iol))):
            if(len(self.Iol) > 0 and not self.Iol.isspace()):
                digitalPinLevelSet.appendChild(Iol)
            else:
                logger.debug("Invalid move, Iol cannot be made")
        if(convNumbIsValid(exNumFromString(self.Ioh))):
            if(len(self.Ioh) > 0 and not self.Ioh.isspace()):
                digitalPinLevelSet.appendChild(Ioh)
            else:
                logger.debug("Invalid move, Ioh cannot be made")
        if(convNumbIsValid(exNumFromString(self.Vcom))):
            if(len(self.Vcom) > 0 and not self.Vcom.isspace()):
                digitalPinLevelSet.appendChild(Vcom)
            else:
                logger.debug("Invalid move, Vcom cannot be made")
        if(convNumbIsValid(exNumFromString(self.Vterm))):
            if(len(self.Vterm) > 0 and not self.Vterm.isspace()):
                digitalPinLevelSet.appendChild(Vterm)
            else:
                logger.debug("Invalid move, Vterm cannot be made")
        if(len(self.TerminationMode) > 0 and not self.TerminationMode.isspace() and (self.TerminationMode == "HighZ" or self.TerminationMode == "Vterm" or self.TerminationMode == "Active Load")):
            digitalPinLevelSet.appendChild(TerminationMode)
        else:
            logger.debug("Invalid move, TerminationMode cannot be made")
        if(len(self.Comment) > 0):
            digitalPinLevelSet.appendChild(Comment)
        else:
            logger.debug("Invalid move, Comment cannot be made")

# Class needed to make the Digital Pin Level Sets
# name - representing the name of the channel
# Vf - Specifies the voltage to force on the specified pin connected to an NI-DCPower instrument.
# If - Specifies the current to force on the specified pin connected to an NI-DCPower instrument.
class DCPowerLevelsSet():
    name = " "
    Vf = " "
    If = " "
    Comment = " "

    #Constructor of this class
    #It sets the value of each attribute, when the instance of the class is made/created
    def __init__(self, name, Vf, If, Comment):
        """Create DC Power level set and set its values

        Args:
            name (string): Level set Name
            Vf (float): Forward Voltage
            If (float): Forward Current
            Comment (string): Comment
        """
        self.name = str(name)
        self.Vf = str(Vf)
        self.If = str(If)
        self.Comment = str(Comment)
        #self.createDCPowerLevelsSets()

    #Function to get the info about components of the pin
    def printPin(self):
        """
            This method will print The Power Pin's name, Vf, If and its comment
        """
        print("PinName  Vf  Vi Comment""\n", self.name, "   |",
              self.Vf, "|",
              self.If, "|",
              self.Comment, " |")
    
    #Function used to create the DC Power Level Sets with the Components from its class.
    #Parameters:
    #@self specifies the instance of the class.
    def createDCPowerLevelsSets(self, root):
        """
        will add a power pin level to the levels sheet
        Args:
            root (PinLevelSheet): the pin level sheet the pin should get appended to
        """
        logger.debug("Adding the PowerPinLevelSet component to the root of the XML doc.")
        DCPowerLevelsSet = root.createElement('PowerPinLevelSet')
        logger.debug("Adding the PowerPinLevelSet component to the root of the XML doc DONE!")
        DCPowerLevelsSet.setAttribute('pin', '%s' % (self.name))
        pinType_ = root.createElement('PowerPinLevelSets')
        root.pinlevelsheetChild.appendChild(pinType_)
        pinType_.appendChild(DCPowerLevelsSet)
        Vf = root.createElement('Vf')
        Vf_atr = root.createTextNode("%s" % (self.Vf))
        Vf.appendChild(Vf_atr)
        If = root.createElement('If')
        If_atr = root.createTextNode("%s" % (self.If))
        If.appendChild(If_atr)
        if(len(self.Vf) > 0 and self.Vf != " "):
            DCPowerLevelsSet.appendChild(Vf)
        else:
            logger.debug("Invalid move, Vf cannot be made")
        if(len(self.If) > 0 and self.If != " "):
            DCPowerLevelsSet.appendChild(If)
        else:
            logger.debug("Invalid move, If cannot be made")  

#Function used to extract the number from the string e.g exNumFromString("3.14") = 3.14
#Parameter:
#@string - specifies a word from which the number is going to be extracted.
def exNumFromString(string):
    """Function used to extract the number from the string e.g exNumFromString("3.14") = 3.14

    Args:
        string (string): specifies a word from which the number is going to be extracted.

    Returns:
        number: the extracted number
    """
    for s in string:
        if s.isdigit():
            return s

#Function to check if the attributes of the classes are the numbers(e.g Vil, Vol...).
##Parameter:
#@number - specifies a number to be checked for its validity. 
def convNumbIsValid(number):
    """Function to check if the attributes of the classes are the numbers(e.g Vil, Vol...)

    Args:
        number (int, float): specifies a number to be checked for its validity

    Returns:
        Bool: 1:valid/0:invalid
    """
    if(number is not None):
        try:
            return True
        except ValueError as val_err:
            logger.debug("Requested Value has to be number! %d"%(val_err))
    else:
        logger.debug("Requested Value has to be number!")

#test function is used only to represent a working example of the whole py file
#Attention should be paid to:
# 1.) Every Channel and its components is used to be represented as
#     string, if we want original and valid digital patern xml digilevel file.
# 2.) 2 Sets of PinLevels: Digital and DCPower
# 3.) The decimal numbers have strict to be written with the ".", example(3.14, 1.41, etc). 
'''
def test():
    p3 = PinLevelSheet()
    p3.digilevelsSourceLogger.setLevel(logging.DEBUG)
    p3.createConsoleHandler()
    p3.digilevelsSourceLogger.addHandler(p3.consoleHandler)
    p3.digilevelsSourceLogger.debug("-----------------STARTING TEST---------------")
    p3.digilevelsSourceLogger.debug("Adding Digital Pin Level Set")
    p3.addDigitalPinLevelSet("CS", "2.5", "0", "2.5", "0.5", "1.5", "-1.5", "0", "0", "HighZ", "HighZ")
    p3.digilevelsSourceLogger.debug("Adding DCPowerLevelSet")
    p3.addDCPowerLevelsSet("DD", "0", "0", "")
    p3.digilevelsSourceLogger.debug("-----------------ENDING TEST------------------")
    p3.digilevelsSourceLogger.debug("Generating file")
    p3.generateXML("test3.digilevels")

if __name__ == "__main__":
    test()
'''


