# author: rste
# date: 25.08.2020
# (c) AMS AG

#this file provides a class to abstract a digipat table and generate+compile a .digipatsrc file to be used with NI digital pattern instruments



import logging
import subprocess
import os

digipatSourceLogger = logging.getLogger(__name__)

#a function to run NI Digital Pattern Compiler from python
#returns the path of the compiled digipat file
def compileDigipatSource(sourcePath, pinmapPath, compilerPath = r"C:\Program Files\National Instruments\Digital Pattern Compiler\DigitalPatternCompiler.exe"):
    """compile a .digipatsrc file into a .digipat file

    Args:
        sourcePath (str): path to the .digipatsrc file
        pinmapPath (str): path to the used .pinmap file
        compilerPath (str, optional): Path to the compiler. Defaults to standard install location: "C:\\Program Files\\National Instruments\\Digital Pattern Compiler\\DigitalPatternCompiler.exe".

    Returns:
        (Bool, str): First tuple element: True if the file was generated by the compiler, False otherwise; Second tuple element: the path to the compiled pattern file
    """
    digipatSourceLogger.debug("running NI Digital Pattern Compiler")
    output = subprocess.Popen([compilerPath, sourcePath, "-pinmap", pinmapPath], stdout=subprocess.PIPE).communicate()[0]
    digipatSourceLogger.debug("NI Digital Pattern Compiler returned:\n" + output.decode("utf-8") )
    return (os.path.exists(sourcePath[:-3]), sourcePath[:-3])

#a class to abstract a vector(line) of the digipatsrc file
class DigipatVector():
    def __init__(self, timeset, drivestates, opcodes = [], label = "", comment = "" ):
        """initialize a digital pattern vector for digipat generation

        Args:
            timeset (str): name of the timeset to use with this vector
            drivestates (list<str>): list of the drivestates for this vector. Keep in mind the order of the Pins!
            opcodes (list, optional): list of opcodes that shall be executed in this vector. Defaults to [].
            label (str, optional): label for this vector. Defaults to "".
            comment (str, optional): optional comment for the file. Defaults to "".
        """
        self.timeset = timeset
        self.label = label
        self.drivestates = drivestates
        self.opcodes = opcodes
        self.comment = comment

#a class to abstract the whole digipatsrc file
class DigipatSource():
    def __init__(self):
        self.usedChannels = []
        self.timesets = set()
        self.vectors = []
        self.exports = set()
        self.imports = set()
        #self.name = getNameFromFileHandler!!!!!!!
        digipatSourceLogger.debug("new DigipatSource Object instantiated")
        pass

    #define which labels to export for other patterns to use
    # *labels: one or more strings containing the names of the labels to export
    # returns: nothing
    def exportLabels(self, *labels):
        """define which labels to export for other patterns to use

        Args:
            *labels: the labels to export
        """

        for label in labels:
            if type(label) == str:
                self.exports.add(label)
                digipatSourceLogger.debug("export label '" + label + "' available in digipattern")

    #define which labels to import from other patterns to use
    # *labels: one or more strings containing the names of the labels to import
    # returns: nothing
    def importLabels(self, *labels):
        """define which labels to import from other patterns to use

        Args:
            *labels: the labels to import
        """
        for label in labels:
            if type(label) == str:
                self.imports.add(label)
                digipatSourceLogger.debug("import label '" + label + "' available in digipattern")

    #define which timesets(defined in the .digitiming file) will be used in the pattern
    # *timesets: one or more strings containing the names of the timesets to be used
    # returns: nothing
    def useTimesets(self, *timesets):
        """define which timesets(defined in the .digitiming file) will be used in the pattern

        Args:
            *timesets: the timesets to use
        """
        for timeset in timesets:
            if type(timeset) == str:
                self.timesets.add(timeset)
                digipatSourceLogger.debug("timeset '" + timeset + "' available in digipattern")

    #define the pins to be used in the digipatsrc file
    # MIND THE ORDER OF THE CHANNELS! the drivestates in each line use the same order furhter on!
    # *channels: one or multiple strings containing the name of the channels to be used
    # returns: nothing
    def useChannels(self, *channels):
        """define the pins to be used in the digipatsrc file
            MIND THE ORDER OF THE CHANNELS! the drivestates in each line use the same order furhter on!
        
        Args:
            *channels: the channels to use
        """
        for channel in channels:
            if type(channel) == str:
                self.usedChannels.append(channel)
                digipatSourceLogger.debug("using channel '" + channel + "'")

    #define the name of the pattern so it can be loaded and used by the nidigital driver
    # patternName: a string containing the name of the pattern
    # returns: nothing
    def setPatternName(self, patternName):
        """define the name of the pattern so it can be loaded and used by the nidigital driver

        Args:
            patternName (str): the name of the pattern
        """
        self.patternName = patternName
        digipatSourceLogger.debug("pattern will be named '" + patternName + "'")

    #add a vector with corresponding opcodes, a timeset and drivestates
    # MIND THE ORDER OF THE DRIVESTATES! the drivestates are evaluated in the same order as the channels are defined in useChannels(...)!
    # timeset: a string containing the name of the timeset to be used with this vector
    # drivestates: an array containing strings with the drivestates the channels should be in in this vector
    # opcodes: an array containing strings with names of the desired opcodes for this vector (default: no opcode)
    # label: string with label of this vector (for flow control with opcodes, default: none)
    # comment: string with comment for this vector
    # returns: nothing
    def addVector(self, timeset, drivestates, opcodes = [], label ="", comment = ""):
        """add a vector with corresponding opcodes, a timeset and drivestates
        MIND THE ORDER OF THE DRIVESTATES! the drivestates are evaluated in the same order as the channels are defined in useChannels(...)

        Args:
            timeset (str): name of the timeset to use with this vector
            drivestates (list<str>): list of the drivestates for this vector. Keep in mind the order of the Pins!
            opcodes (list, optional): list of opcodes that shall be executed in this vector. Defaults to [].
            label (str, optional): label for this vector. Defaults to "".
            comment (str, optional): optional comment for the file. Defaults to "".
        """
        vector = DigipatVector(timeset, drivestates, opcodes, label, comment)
        #check if timeset exists
        if not vector.timeset in self.timesets:
            digipatSourceLogger.debug("timeset '" + str(timeset) + "' not available in digipattern, adding it automatically")
            self.timesets.add(vector.timeset)
        self.vectors.append(vector)
        digipatSourceLogger.debug("added vector: " + str(vector.opcodes) + " " + str(vector.timeset) + " " + str(vector.drivestates) )

    #generate a .digipatsrc file from the information provided by set/use/add functions
    # filepath: the path to where the file should be saved. Empty string = dont save the file
    # returns: the textual content of the file
    def generateDigipatsrcFile(self, filepath = ""):
        """Generate the .digipatsrc content

        Args:
            filepath (str): Path/Name of a file to save the .digipatsrc to. Not saving if left empty. Defaults to "".

        Returns:
            str: the .digipatsrc content
        """
        #build file content
        #header
        content = "\nfile_format_version 1.1;\n"

        #exports
        if bool(self.exports): #if exports are not empty
            content += "export "
            for export in self.exports:
                content += export + ", "
            #swap last comma + blank with semicolon and newline
            content = content[:-2] + ";\n"

        #imports
        if bool(self.imports): #if imports are not empty
            content += "import "
            for _import in self.imports:
                content += _import + ", "
            #swap last comma + blank with semicolon and newline
            content = content[:-2] + ";\n"

        #timesets
        content += "timeset "
        for timeset in self.timesets:
            content += timeset + ", "
        #swap last comma + blank with semicolon and newline
        content = content[:-2] + ";\n"
        content += "\n"

        #pattern
        #pattern header
        content += "pattern " + self.patternName + " ("
        #add all used channels
        for channel in self.usedChannels:
            content += channel + ", "
        #close the bracket, remove last comma
        content = content[:-2] + ")\n"

        #pattern block with vectors
        content += "{\n"
        for vector in self.vectors:
            vectorString = ""
            #label
            if vector.label != "":
                vectorString += vector.label + ":\t"

            #opcodes
            opcodeString = ""
            for opcode in vector.opcodes:
                opcodeString += opcode + ", "
            vectorString += opcodeString[0:-2] #omit last ", "
            vectorString += "\t"

            #timeset
            vectorString += vector.timeset + "\t"

            #drivestates
            drivestateString = ""
            for drivestate in vector.drivestates:
                drivestateString += str(drivestate) + "\t" #typecast to string, so drivestate can be 1 or 0 as integer too
            vectorString += drivestateString[0:-1] #omit last "\t"

            #semicolon at the end of the vector
            vectorString += ";"

            #comment at the very end
            if vector.comment != "":
                vectorString += "//" + vector.comment

            #add each vector
            content += vectorString +"\n"

        #close pattern block
        content += "}\n"

        digipatSourceLogger.debug("generated file content: \n" + content)

        #save if necessary
        if filepath != "" and filepath[-11:] == ".digipatsrc":
            digipatSourceLogger.info("writing digipatsrc file: " + filepath)

            with open(filepath, "w") as digipatsrcFile:
                digipatsrcFile.write(content)

        #return the content
        return content
