# author: rste
# date: 26.08.2020
# (c) AMS AG

#this file provides classes to abstract a digitiming Time Sheet from the NI digital Pattern Editor and save it as .digitiming file (xml with fancy file ending)

import logging
from xml.dom import minidom, Node

digitimingLogger = logging.getLogger(__name__)

#class to define the behaviour of a pin within a TimeSet
#for further information on what the parameters do, please visit: http://zone.ni.com/reference/en-XX/help/375145e/digipat/timingtwo/
# EDGE MULTIPLIERS ARE NOT SUPPORTED!
# pinItem: string
# driveFormat: string, "RH", "RL", "NR" or "SBC"
# driveOn: float/int or string with formula
# driveData: float/int or string with formula
# driveReturn: float/int or string with formula
# driveOff: float/int or string with formula
# compareStrobe: float/int or string with formula
# comment: string
class PinEdge():
    def __init__(self, pinItem, driveFormat, driveOn, driveData, driveReturn, driveOff, compareStrobe, comment=""):
        """[summary]

        Args:
            pinItem (str): name of the Pin/PinGroup
            driveFormat (str): drive format. Possible values: "ReturnToHigh"("RH"), "ReturnToLow"("RL"), "DriveNonReturn"("NR"), "SurroundByComplement"("SBC")
            driveOn (float): Specifies the delay from the beginning of the vector period for turning on the pin driver. This option applies only when the previous vector left the pin in a non-drive pin state (L, H, X, V, M, E)
            driveData (float): Specifies the delay from the beginning of the vector period until the pattern data is driven to the pattern value.
            driveReturn (float): Specifies the delay from the beginning of the vector period until the pin changes from the pattern data to the return value, as specified by the format
            driveOff (float): Specifies the delay from the beginning of the vector period to turn off the pin driver when the next vector period uses a non-drive symbol (L, H, X, V, M, E). 
            compareStrobe (float): Specifies the time when the comparison happens within a vector period. 
            comment (str, optional): Optional Comment for the digitiming file. Defaults to "".
        """
        if (type(pinItem) == str):
            self.pinItem = pinItem
        else:
            digitimingLogger.error("Parameter 'pinItem' in PinEdge must be str! parameter was: "+ str( type(pinItem) ) )

        driveFormatU = driveFormat.upper()
        if (driveFormatU == "RH" or driveFormatU == "RETURNTOHIGH"):
            self.driveFormat = "RH"
        elif (driveFormatU == "RL" or driveFormatU == "RETURNTOLOW"):
            self.driveFormat = "RL"
        elif (driveFormatU == "NR" or driveFormatU == "DRIVENONRETURN"):
            self.driveFormat = "NR"
        elif (driveFormatU == "SBC" or driveFormatU == "SURROUNDBYCOMPLEMENT"):
            self.driveFormat = "SBC"
        else:
            digitimingLogger.error("Parameter 'driveFormat' in PinEdge must be str! ('RH'/'RETURNTOHIGH', 'RL'/'RETURNTOLOW', 'NR'/'DRIVENONRETURN' or 'SBC'/'SURROUNDBYCOMPLEMENT') Was: " + str(driveFormat) + "(" + str(type(driveFormat)) + ")")

        if (type(driveOn) == str or type(driveOn) == float or type(driveOn) == int):
            self.driveOn = driveOn
        else:
            digitimingLogger.error("Parameter 'driveOn' in PinEdge must be either str, float or int! Was: " + str(type(driveOn)))

        if (type(driveData) == str or type(driveData) == float or type(driveData) == int):
            self.driveData = driveData
        else:
            digitimingLogger.error("Parameter 'driveData' in PinEdge must be either str, float or int! Was: " + str(type(driveData)))

        if (type(driveReturn) == str or type(driveReturn) == float or type(driveReturn) == int):
            self.driveReturn = driveReturn
        else:
            digitimingLogger.error("Parameter 'driveReturn' in PinEdge must be either str, float or int! Was: " + str(type(driveReturn)) )

        if (type(driveOff) == str or type(driveOff) == float or type(driveOff) == int):
            self.driveOff = driveOff
        else:
            digitimingLogger.error("Parameter 'driveOff' in PinEdge must be either str, float or int! Was: " + str(type(driveOff)) )

        if (type(compareStrobe) == str or type(compareStrobe) == float or type(compareStrobe) == int):
            self.compareStrobe = compareStrobe
        else:
            digitimingLogger.error("Parameter 'compareStrobe' in PinEdge must be either str, float or int! Was: " + str(type(compareStrobe)) )

        if (type(comment) == str):
            self.comment = comment
        else:
            digitimingLogger.error("Parameter 'comment' in PinEdge must be str!Was: " + str(type(comment)) )

        digitimingLogger.debug("creating PinEdge: " + pinItem + ", " + driveFormat + ", " + str(driveOn) + ", " + str(driveData) + ", " + str(driveReturn) + ", " + str(driveOff) + ", " + str(compareStrobe) + ", " + comment)

# a class to build up a TimeSet
# name: string, name of the time set
# period: float or string containing formula
class TimeSet():
    def __init__(self, name, period):
        """Initialize a Time Set to be used by the digital pattern instrument

        Args:
            name (str): The name of the timeset
            period (float): The period of the timeset
        """
        if (type(name) == str):
            self.name = name
        else:
            digitimingLogger.error("Parameter 'name' in PinEdge must be str! Was: " + str(type(name)))

        if (type(period) == str or type(period) == float or type(period) == int):
            self.period = period
        else:
            digitimingLogger.error("Parameter 'period' in PinEdge must be either str, float or int! Was: " + str(type(period)) )

        self.pinEdges = []

        digitimingLogger.debug("creating TimeSet: " + name + ", " + str(period))

    #add a PinEdge to the timeset
    # MIND THIS! all pins defined in the pinmap should have a pinEdge for each time set
    # pinEdge: PinEdge object
    # returns: nothing
    def addPinEdge(self, pinEdge):
        """add a PinEdge to the timeset

        Args:
            pinEdge (PinEdge): The PinEdge Description to add to this timeset
        """
        self.pinEdges.append(pinEdge)

class Digitiming:
    def __init__(self):
        self.timesets = []
        pass

    # add a timeset containing all timings for pins (pinEdges)
    # timeset: a TimeSet object
    # returns: nothing
    def addTimeSet(self, timeset):
        """add a timeset to the digitiming document

        Args:
            timeset (TimeSet): The timeset to be added to the digitiming document
        """
        self.timesets.append(timeset)

    # generate .digitiming file from all the set Timesets
    # filepath: string with the filepath where the xml (.digitiming) file shall be saved. If left empty/file ending is not .digitiming, the file will not be saved
    # returns: the xml content of the file as string
    def generateDigitimingXML(self, filepath = ""):
        """generate the xml content for the digitiming

        Args:
            filepath (str, optional): path/name of the digitiming file to be saved. Not saving if left empty. Defaults to "".

        Returns:
            str: the digitiming XML content
        """

        #build the xml document
        doc = minidom.Document()

        #function for simplifying creation of elements with text content
        def createTextElement(name, text):
            element = doc.createElement(name)
            elementText = doc.createTextNode( str(text) )
            element.appendChild(elementText)
            return element

        #header stuff
        timingFile = doc.createElement("TimingFile")
        timingFile.setAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema")
        timingFile.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        timingFile.setAttribute("schemaVersion", "1.0")
        timingFile.setAttribute("xmlns", "http://www.ni.com/Semiconductor/Timing")
        doc.appendChild(timingFile)
        timingSheet = doc.createElement("TimingSheet")
        timingFile.appendChild(timingSheet)

        #time sets
        timeSets = doc.createElement("TimeSets")
        timingSheet.appendChild(timeSets)

        #domTimeset is the xml object, timeset is the python abstraction
        for timeset in self.timesets:
            domTimeSet = doc.createElement("TimeSet")
            domTimeSet.setAttribute("name", timeset.name)
            timeSets.appendChild(domTimeSet)

            domPeriod = createTextElement("Period", timeset.period)
            domTimeSet.appendChild(domPeriod)

            #PinEdges within timeset
            domPinEdges = doc.createElement("PinEdges")
            domTimeSet.appendChild(domPinEdges)

            for pinEdge in timeset.pinEdges:
                domPinEdge = doc.createElement("PinEdge")
                domPinEdge.setAttribute("pin", pinEdge.pinItem)
                domPinEdges.appendChild(domPinEdge)

                #non return pin edge
                if pinEdge.driveFormat == "NR":
                    domDriveNonReturn = doc.createElement("DriveNonReturn")
                    domPinEdge.appendChild(domDriveNonReturn)
                    domOn = createTextElement("On", pinEdge.driveOn)
                    domData = createTextElement("Data", pinEdge.driveData)
                    domOff = createTextElement("Off", pinEdge.driveOff)
                    domDriveNonReturn.appendChild(domOn)
                    domDriveNonReturn.appendChild(domData)
                    domDriveNonReturn.appendChild(domOff)

                #return pin edges
                if (pinEdge.driveFormat == "RH" or pinEdge.driveFormat == "RL" or pinEdge.driveFormat == "SBC"):
                    if pinEdge.driveFormat == "RH":
                        domDriveFormat = doc.createElement("ReturnToHigh")
                    elif pinEdge.driveFormat == "RL":
                        domDriveFormat = doc.createElement("ReturnToLow")
                    else:
                        domDriveFormat = doc.createElement("SurroundByComplement")

                    domPinEdge.appendChild(domDriveFormat)

                    domOn = createTextElement("On", pinEdge.driveOn)
                    domData = createTextElement("Data", pinEdge.driveData)
                    domReturn = createTextElement("Return", pinEdge.driveReturn)
                    domOff = createTextElement("Off", pinEdge.driveOff)
                    domDriveFormat.appendChild(domOn)
                    domDriveFormat.appendChild(domData)
                    domDriveFormat.appendChild(domReturn)
                    domDriveFormat.appendChild(domOff)

                #comparator strobe timing
                domCompareStrobe = doc.createElement("CompareStrobe")
                domPinEdge.appendChild(domCompareStrobe)
                domStrobe = createTextElement("Strobe", pinEdge.compareStrobe)
                domCompareStrobe.appendChild(domStrobe)

                #data source? - not clear what this element is for
                domPinEdge.appendChild(createTextElement("DataSource", "Pattern"))
                #comment
                domPinEdge.appendChild(createTextElement("Comment", pinEdge.comment))

        #generate XML from the built DOM
        xmlString = doc.toprettyxml(indent = '  ', encoding = "utf-8").decode("utf-8")

        digitimingLogger.debug("generated digitiming xml content:\n" + xmlString)

        #save xml if desired
        if filepath != "" and filepath[-11:] == ".digitiming":
            digitimingLogger.info("writing digitiming file: " + filepath)
            with open(filepath, "w") as digitimingFile:
                digitimingFile.write(xmlString)

        return xmlString
