# author: rste
# date: 26.08.2020
# (c) AMS AG

#this file provides classes to abstract a specs Specification sheet from the NI digital Pattern Editor and save it as .specs file (xml with fancy file ending)

import logging
from xml.dom import minidom, Node

specsLogger = logging.getLogger("specs")

class Variable():
    def __init__(self, symbol, formula, comment= ""):
        """initialize nidigital specs variable

        Args:
            symbol (str): symbolic name of the variable
            formula (str/float/int): formula to calculate the variable or value
            comment (str, optional): [description]. Defaults to "".
        """
        if (type(symbol) == str):
            self.symbol = symbol
        else:
            specsLogger.error("Parameter 'symbol' in Section must be str! parameter was: "+ str( type(symbol) ) )

        if (type(formula) == str or type(formula) == float or type(formula) == int):
            self.formula = formula
        else:
            specsLogger.error("Parameter 'formula' in Section must be either str, float or int! parameter was: "+ str( type(formula) ) )

        if (type(comment) == str):
            self.comment = comment
        else:
            specsLogger.error("Parameter 'comment' in Section must be str! parameter was: "+ str( type(comment) ) )

        specsLogger.debug("create Variable: " + symbol + ", " + str(formula) + ", " + comment)

#A section for variables. Each Variable must belong to a section for some reason (courtesy of National Instruments)
class Section():
    def __init__(self, name):
        """Initialize a variable section

        Args:
            name (str): name of the section
        """
        if (type(name) == str):
            self.name = name
        else:
            specsLogger.error("Parameter 'name' in Section must be str! parameter was: "+ str( type(name) ) )

        self.variables = []
        specsLogger.debug("Creating section: " + name)

    # add a variable to this section
    # variable: a Variable Object
    # returns: nothing
    def addVariable(self, variable):
        """add a variable to this section

        Args:
            variable (Variable): variable to add
        """
        self.variables.append(variable)

class Specifications():
    def __init__(self):
        """initialize specifications object
        """
        self.sections = []

    # add a section containing variables to the specifications sheet
    # section: a Section Object
    # returns: nothing
    def addSection(self, section):
        """add a section of variables to the Specification

        Args:
            section (Section): variable section to add
        """
        self.sections.append(section)

    # generate .specs file from all the set Sections/Variables
    # filepath: string with the filepath where the xml (.specs) file shall be saved. If left empty/file ending is not .specs, the file will not be saved
    # returns: the xml content of the file as string
    def generateSpecsXML(self, filepath = ""):
        """generate .specs file from all the set Sections/Variables
        Args:
            filepath (str, optional): path/name of the .specs file to save. Not saving if left empty. Defaults to "".

        Returns:
            str: the .specs xml content
        """

        #build the xml document
        doc = minidom.Document()

        #function for simplifying creation of elements with text content
        def createTextElement(name, text):
            element = doc.createElement(name)
            elementText = doc.createTextNode( str(text) )
            element.appendChild(elementText)
            return element

        #header stuff
        domSpecifications = doc.createElement("Specifications")
        domSpecifications.setAttribute("xmlns:f", "http://www.ni.com/schemas/Semiconductor/Formula.xsd")
        domSpecifications.setAttribute("schemaVersion", "1.0")
        domSpecifications.setAttribute("xmlns", "http://www.ni.com/schemas/Semiconductor/Specifications.xsd")
        doc.appendChild(domSpecifications)

        #sections
        for section in self.sections:
            domSection = doc.createElement("Section")
            domSection.setAttribute("name", section.name)
            domSpecifications.appendChild(domSection)

            #variables in the section
            for variable in section.variables:
                domFormula = doc.createElement("f:Formula")
                domFormula.setAttribute("symbol", variable.symbol)
                domSection.appendChild(domFormula)

                domFormula.appendChild( createTextElement("f:Definition", variable.formula) )
                if (variable.comment):
                    domFormula.appendChild( createTextElement("f:Comment", variable.comment) )

        #generate XML from the built DOM
        xmlString = doc.toprettyxml(indent = '  ', encoding = "utf-8").decode("utf-8")

        specsLogger.debug("generated specs xml content:\n" + xmlString)

        #save xml if desired
        if filepath != "" and filepath[-6:] == ".specs":
            specsLogger.info("writing specs file: " + filepath)
            with open(filepath, "w") as specsFile:
                specsFile.write(xmlString)